-- debug_status = 1
debug_mod_name = "TimeTools"
debug_file = debug_mod_name .. "-debug.txt"
require("utils")
require("config")

local ticks_per_day = 25000

--------------------------------------------------------------------------------------
local function approx_day()
	global.day = 1+math.floor((game.tick+12500) / ticks_per_day)
end

--------------------------------------------------------------------------------------
function init_globals()	
	debug_print( "init globals" )
	
	global.ticks = global.ticks or 0
	global.cycles = global.cycles or 0
	global.day = global.day or 0
	global.frozen = global.frozen or false
	global.clocks = global.clocks or {}
	global.h_prev = global.h_prev or 23
	global.speed_mem = global.speed_mem or maximum_speed
	global.always_day = global.always_day or false
	global.surface = game.surfaces.nauvis
end

--------------------------------------------------------------------------------------
script.on_init( 
	function()
		debug_print( "on_init" )
		init_globals()
	end
)

--------------------------------------------------------------------------------------
function on_configuration_changed(data)
	-- detect any mod or game version change
	if data.mod_changes ~= nil then
		local changes = data.mod_changes[debug_mod_name]
		if changes ~= nil then
			debug_print( "update mod: ", debug_mod_name, " ", tostring(changes.old_version), " to ", tostring(changes.new_version) )
		
			init_globals()
			
			global.always_day = -1 -- to force update of icon
			
			for _,force in pairs(game.forces) do
				force.reset_recipes()
				force.reset_technologies()
				force.recipes["clock-combinator"].enabled = force.technologies["circuit-network"].researched
			end
			
			for _, player in pairs(game.players) do
				if player.gui.top.timebar_frame then player.gui.top.timebar_frame.destroy() end -- destroy old bar
				if player.gui.top.timetools_frame then player.gui.top.timetools_frame.destroy() end -- destroy new bar to refresh it
				build_gui(player)
			end

			update_guis()			
		end
	end
end

script.on_configuration_changed(on_configuration_changed)

--------------------------------------------------------------------------------------
function on_player_created(event)
	debug_print( "player created" )
	local player = game.players[event.player_index]
	
	build_gui(player)
end

script.on_event(defines.events.on_player_created, on_player_created )

--------------------------------------------------------------------------------------
function on_creation( event )
	local ent = event.created_entity
	
	if ent.name == "clock-combinator" then
		debug_print( "clock-combinator created" )
		
		table.insert( global.clocks, 
			{
				entity = ent, 
			}
		)
		
		debug_print( "clocks=" .. #global.clocks )
	end
end

script.on_event(defines.events.on_built_entity, on_creation )
script.on_event(defines.events.on_robot_built_entity, on_creation )

--------------------------------------------------------------------------------------
function on_destruction( event )
	local ent = event.entity
	
	if ent.name == "clock-combinator" then
		debug_print( "clock-combinator destroyed" )
		
		for i, clock in ipairs(global.clocks) do
			if clock.entity == ent then
				table.remove( global.clocks, i )
				break
			end
		end
		
		debug_print( "clocks=" .. #global.clocks )
	end
end

script.on_event(defines.events.on_entity_died, on_destruction )
script.on_event(defines.events.on_robot_pre_mined, on_destruction )
script.on_event(defines.events.on_preplayer_mined_item, on_destruction )

--------------------------------------------------------------------------------------
function on_tick(event)
	if global.ticks <= 0 then
		global.ticks = clock_cycle_in_ticks

		-- daytime : 0.0 to 1.0, noon to noon (midnight at 0.5), max light to min light to max light...
		-- game starts at daytime = 0, so noon of day 1.
		tim = game.surfaces.nauvis.daytime
		tim = (tim*24+12) % 24
		tim_h = math.floor(tim)
		tim_m = math.floor((tim-tim_h)*60)
		
		if game.surfaces.nauvis.always_day then
			approx_day()
		else
			if tim_h == 0 and global.h_prev == 23 then
				global.day = global.day + 1
			end
		end

		-- update time display on button
		
		local s_time = string.format("%u-%02u:%02u", global.day, tim_h, tim_m )
		local flow
		local refresh_always_day = (game.surfaces.nauvis.always_day ~= global.always_day)
		
		for _, player in pairs(game.players) do
			if player.connected and player.gui.top.timetools_frame then
				flow = player.gui.top.timetools_frame.timetools_flow
				flow.but_time.caption = s_time
				
				if refresh_always_day then
					if game.surfaces.nauvis.always_day then
						flow.but_always.style = "timetools_button_always_day_style"
					else
						flow.but_always.style = "timetools_button_night_and_day_style"
					end
				end
				if debug_status then
					flow.but_tick.caption = game.tick
				end
			end
		end

		global.always_day = game.surfaces.nauvis.always_day
		global.h_prev = tim_h

		if global.cycles <= 0 then
			global.cycles = clock_combinator_cycle
			-- update clocks variables
			for i, clock in pairs(global.clocks) do
				if clock.entity.valid then
					params = {parameters={
						{index=1,signal={type="virtual",name="signal-clock-gametick"},count=math.floor(game.tick)},
						{index=2,signal={type="virtual",name="signal-clock-day"},count=global.day},
						{index=3,signal={type="virtual",name="signal-clock-hour"},count=tim_h},
						{index=4,signal={type="virtual",name="signal-clock-minute"},count=tim_m},
						{index=5,signal={type="virtual",name="signal-clock-alwaysday"},count=iif(game.surfaces.nauvis.always_day,1,0)},
						{index=6,signal={type="virtual",name="signal-clock-darkness"},count=math.floor(game.surfaces.nauvis.darkness*100)},
						{index=7,signal={type="virtual",name="signal-clock-lightness"},count=math.floor((1-game.surfaces.nauvis.darkness)*100)},
					}}
					
					-- problem here to solve
					-- clock.entity.set_circuit_condition( defines.circuitconditionindex.constant_combinator, params )
					clock.entity.get_control_behavior().parameters = params
				else
					table.remove(global.clocks,i)
				end
			end
		else
			global.cycles = global.cycles - 1
		end
	else
		global.ticks = global.ticks - 1
	end
end

script.on_event(defines.events.on_tick, on_tick)

--------------------------------------------------------------------------------------
function on_gui_click(event)
	if event.element.name == "but_time" then
		if game.surfaces.nauvis.always_day then
			game.surfaces.nauvis.always_day = false
			global.frozen = false
		else
			global.frozen = not global.frozen
		end
		game.surfaces.nauvis.freeze_daytime( global.frozen )
		update_guis()
		
	elseif event.element.name == "but_always" then
		game.surfaces.nauvis.always_day = not game.surfaces.nauvis.always_day
		
		if game.surfaces.nauvis.always_day then
			global.frozen = false
		end
		update_guis()
		
	elseif event.element.name == "but_slower" then
		if game.speed >= 0.2 then game.speed = game.speed / 2 end -- minimum 0.1
		if game.speed ~= 1 then global.speed_mem = game.speed end
		update_guis()
		
	elseif event.element.name == "but_faster" then
		if game.speed < maximum_speed then game.speed = game.speed * 2 end
		if game.speed ~= 1 then global.speed_mem = game.speed end
		update_guis()

	elseif event.element.name == "but_speed" then
		if game.speed == 1 then game.speed = global.speed_mem else game.speed = 1 end
		update_guis()
	end
	
end

script.on_event(defines.events.on_gui_click, on_gui_click )

--------------------------------------------------------------------------------------
function build_gui( player )
	local guif = player.gui.top.timetools_frame
	
	if guif == nil then
		debug_print("create frame player" .. player.name)
		guif = player.gui.top.add({type = "frame", name = "timetools_frame", caption = "", direction = "horizontal", style = "timetools_frame_style"})
		guif = guif.add({type = "flow", name = "timetools_flow", direction = "horizontal", style = "timetools_flow_style"})
		guif.add({type = "button", name = "but_time", caption = "0-00:00", font_color = colors.white, style = "timetools_button_style"})				
		guif.add({type = "button", name = "but_always", style = "timetools_button_night_and_day_style"})
		guif.add({type = "button", name = "but_slower", caption = "<" , font_color = colors.white, style = "timetools_button_style"})
		guif.add({type = "button", name = "but_speed", caption = "x1" , font_color = colors.white, style = "timetools_button_style"})
		guif.add({type = "button", name = "but_faster", caption = ">" , font_color = colors.white, style = "timetools_button_style"})
		if debug_status then
			guif.add({type = "button", name = "but_tick", caption = "0" , font_color = colors.white, style = "timetools_button_style"})
		end
		return( guif )
	else
		return( guif.timetools_flow )
	end
	
end
	
--------------------------------------------------------------------------------------
function update_guis()
	local display_red = (global.frozen or game.surfaces.nauvis.always_day)
	for _, player in pairs(game.players) do
		if player.connected then
			local flow = build_gui(player)
			local s
			
			if game.speed == 1 then
				flow.but_speed.caption = "x1"
				flow.but_speed.style.font_color = colors.white
			elseif game.speed < 1 then
				s = string.format("/%1.0f", 1/game.speed )
				flow.but_speed.caption = s
				flow.but_speed.style.font_color = colors.green
			elseif game.speed > 1 then
				s = string.format("x%1.0f", game.speed )
				flow.but_speed.caption = s
				flow.but_speed.style.font_color = colors.lightred
			end
	
			if game.surfaces.nauvis.always_day then
				flow.but_always.style = "timetools_button_always_day_style"
			else
				flow.but_always.style = "timetools_button_night_and_day_style"
			end

			if display_red then
				flow.but_time.style.font_color = colors.lightred
			else
				flow.but_time.style.font_color = colors.green
			end
		end
	end	
end

--------------------------------------------------------------------------------------

local interface = {}

function interface.setfrozen( frozen )
	debug_print( "frozen" )
	
	global.frozen = frozen
	
	update_guis()
end

function interface.setday( dd )
	debug_print( "setday" )
	
	if dd == nil then dd = 0 end
	
	global.day = math.floor(dd)
	
	update_guis()
end

function interface.setclock( hhmm )
	debug_print( "setclock" )
	
	if hhmm == nil then hhmm = 0 end
	if hhmm < 0 then hhmm = 0 end
	if hhmm >= 24 then hhmm = 23.59 end
	
	local mm = (hhmm - math.floor(hhmm)) * 100
	local hh = math.floor(hhmm)
	
	if mm >= 60 then mm = 59 end
	
	if hh < 12 then
		hh = hh + 12
	else
		hh = hh - 12
	end
	
	game.surfaces.nauvis.daytime = min((hh * 60 + mm) / 24 / 60,1)
	game.surfaces.nauvis.always_day = false
	
	update_guis()
end

function interface.reset()
	debug_print( "reset" )
	
	approx_day()
	
	for _,force in pairs(game.forces) do
		force.reset_recipes()
		force.reset_technologies()
	end
	
	for _, player in pairs(game.players) do
		if player.gui.top.timetools_frame then	
			player.gui.top.timetools_frame.destroy()
		end
	end
	
	update_guis()
end

remote.add_interface( "timetools", interface )

-- /c remote.call( "timetools", "setday", 9 )
-- /c remote.call( "timetools", "setclock", 20.15 )
-- /c remote.call( "timetools", "reset" )
