require("defines")
require("util")
require("config")

-- Constants
-- The position shift for item output for each direction.
local item_source_shift =
{
	-- The values are shifted from the original mod so surface.can_place_entity can work properly.
	[defines.direction.north] = {x1 = 0.3, y1 = 0.9, x2 = -0.3, y2 = 0.9},
	[defines.direction.east] = {x1 = -0.9, y1 = 0.3, x2 = -0.9, y2 = -0.3},
	[defines.direction.south] = {x1 = -0.3, y1 = -0.9, x2 = 0.3, y2 = -0.9},
	[defines.direction.west] = {x1 = 0.9, y1 = -0.3, x2 = 0.9, y2 = 0.3},
}
-- The position shift for item removal for each direction.
local item_void_shift =
{
	[defines.direction.north] = {x = 0, y = -0.9},
	[defines.direction.east] = {x = 0.9, y = 0},
	[defines.direction.south] = {x = 0, y = 0.9},
	[defines.direction.west] = {x = -0.9, y = 0}
}

-- Variables
-- The event ID for the creative mode on-enabled event.
local on_enabled_event_id

-- Functions
-- Returns the event ID for the creative mode on-enabled event. If it has not been loaded, it will be loaded first.
function get_or_load_on_enabled_event_id()
	if on_enabled_event_id == nil then
		on_enabled_event_id = script.generate_event_name()
	end
	return on_enabled_event_id
end

-- Returns the bounding box of the grid at the given position.
function get_tile_bb(position)
	return
	{
		{math.floor(position.x), math.floor(position.y)},
		{math.ceil(position.x), math.ceil(position.y)}
	}
end

-- Returns whether the given entity has an inventory.
function has_inventory(entity)
	for _, inv in pairs(defines.inventory) do
		if entity.get_inventory(inv) ~= nil then
			return true
		end
	end
	return false
end

-- Returns whether the given table contain the given specific value.
function has_value(tab, val)
	for _, v in ipairs(tab) do
		if v == val then
			return true
		end
	end
	return false
end

-- Removes the first item on the given transport line.
function remove_first_item_on_transport_line(line)
	local contents = line.get_contents()
	while not line.can_insert_at(0) do
		local item = next(contents)
		local stack = {name = item, count = 1}
		line.remove_item(stack)
	end
end

-- Removes one item from an inventory of the given entity.
function remove_one_item_in_entity(entity)
	for _, inv in pairs(defines.inventory) do
		local inventory = entity.get_inventory(inv)
		if inventory then
			if not inventory.is_empty() then
				local contents = inventory.get_contents()
				local item = next(contents)
				local stack = {name = item, count = 1}
				inventory.remove(stack)
				return
			end
		end
	end
end

-- Outputs the item of given name at the given position or remove the items there.
-- Returns the updated list of player entities in front of it, so it should not insert the items to them in the next tick.
function output_or_remove_item(surface, position, shift_x, shift_y, direction, is_output_item, output_item_name, output_item_slot, output_slot_inserted_players)
	-- Create a simple item stack according to the given item name of we are going to output it.
	local output_stack
	if is_output_item then
		output_stack = {name = output_item_name, count = 1}
	end
	-- Add the offset to the position to get the actual one.
	local actual_position = {x = position.x + shift_x, y = position.y + shift_y}
	-- Store the entities to insert items to or remove items from them with different priorities.
	local other_entity = nil
	local train_entity = nil
	-- There may be multiple cars and players in front of it, so we will use table to store them.
	local car_entities = nil
	local player_entities = nil
	-- Whether the item should be dropped on ground as the last resort.
	local drop_on_ground = is_output_item and creative_mode_config.item_source.can_drop_on_ground
	-- Whether the items on ground should be removed.
	local remove_from_ground = not is_output_item and creative_mode_config.item_void.can_remove_from_ground
	
	-- Check if there is any entity in the grid at the given position.
	local tile_bb = get_tile_bb(actual_position)
	for _, entity in ipairs(surface.find_entities(tile_bb)) do
		-- If the entity is a transport-belt, simply make effect on it because it has the highest priority.
		local isBelt = entity.type == "transport-belt"
		local isUndergroundBelt = not isBelt and entity.type == "underground-belt"
		if isBelt or isUndergroundBelt then
			if is_output_item then
				-- Output item only if the transport's direction matches.
				if entity.direction == util.oppositedirection(direction) then
					-- For underground belt, it also needs to be input type.
					if isBelt or (isUndergroundBelt and entity.belt_to_ground_type == "input") then
						-- Insert the item at the back of the transport line if possible.
						local line_num
						if output_item_slot == 1 then
							line_num = defines.transport_line.left_line
						else
							line_num = defines.transport_line.right_line
						end
						local transport_line = entity.get_transport_line(line_num)
						if transport_line.can_insert_at_back() then
							transport_line.insert_at_back(output_stack)
						end
					end
				end
			else
				if entity.direction == util.oppositedirection(direction) then
					if isBelt or (isUndergroundBelt and entity.belt_to_ground_type == "output") then
						-- Remove the first item on each of the transport lines.
						remove_first_item_on_transport_line(entity.get_transport_line(defines.transport_line.left_line))
						remove_first_item_on_transport_line(entity.get_transport_line(defines.transport_line.right_line))
					end
				end
			end
			-- No need to check for another entity.
			return nil
		elseif entity.type == "locomotive" or entity.type == "cargo-wagon" then
			-- The entity is a train, check config.
			if (is_output_item and creative_mode_config.item_source.can_insert_to_vehicle) or (not is_output_item and creative_mode_config.item_void.can_remove_from_vehicle) then
				train_entity = entity
			end
		elseif entity.type == "car" then
			-- The entity is a car, check config.
			if (is_output_item and creative_mode_config.item_source.can_insert_to_vehicle) or (not is_output_item and creative_mode_config.item_void.can_remove_from_vehicle) then
				if not car_entities then car_entities = {} end
				table.insert(car_entities, entity)
			end
		elseif entity.type == "player" then
			-- The entity is a player, we will need to do more checkings from the config.
			if (is_output_item and creative_mode_config.item_source.can_insert_to_player) or (not is_output_item and creative_mode_config.item_void.can_remove_from_player) then
				if not player_entities then player_entities = {} end
				table.insert(player_entities, entity)
			end
		else
			-- For other entity types, we have to check whether they have inventory so that unwanted entities like particles are ignored.
			if has_inventory(entity) then
				other_entity = entity
			end
		end
	end
	
	-- Actually output the item or remove the items according to the found entities with different priorities.
	if other_entity then
		-- Other entities that are not train, car nor player.
		if is_output_item then
			if other_entity.can_insert(output_stack) then
				other_entity.insert(output_stack)
			end
		else
			remove_one_item_in_entity(other_entity)
		end
		return nil
	elseif train_entity then
		-- Train. Output or remove only if the train is stationary.
		if train_entity.train.speed == 0 then
			if is_output_item then
				if train_entity.can_insert(output_stack) then
					train_entity.insert(output_stack)
				end
			else
				remove_one_item_in_entity(train_entity)
			end
		end
		return nil
	elseif car_entities then
		-- Cars. Output or remove only if the car is stationary.
		for _, car in ipairs(car_entities) do
			if car.speed == 0 then
				if is_output_item then
					if car.can_insert(output_stack) then
						car.insert(output_stack)
					end
				else
					remove_one_item_in_entity(car)
				end
			end
		end
		return nil
	elseif player_entities then
		-- Players.
		if is_output_item then
			-- Change the item count if needed.
			if creative_mode_config.item_source.insert_only_once_to_player then
				if creative_mode_config.item_source.insert_to_player_by_stack then
					output_stack.count = game.item_prototypes[output_stack.name].stack_size * creative_mode_config.item_source.insert_to_player_amount
				else
					output_stack.count = creative_mode_config.item_source.insert_to_player_amount
				end
			end
			for _, player in ipairs(player_entities) do
				-- Check if the player is restricted to get the item only once so the item cannot be inserted this time.
				if not creative_mode_config.item_source.insert_only_once_to_player or (not output_slot_inserted_players or not has_value(output_slot_inserted_players, player)) then
					if player.can_insert(output_stack) then
						player.insert(output_stack)
					end
				end
			end
			-- Returns the table of player entities to make sure they are not inserted the item if config doesn't allow to.
			return player_entities
		else
			for _, player in ipairs(player_entities) do
				remove_one_item_in_entity(player)
			end
			return nil
		end
	end
	
	-- The last resort. Drop item on ground or remove items on ground.
	if drop_on_ground then
		if surface.can_place_entity{name = "item-on-ground", position = actual_position, stack = item} then
			surface.create_entity
			{
				name = "item-on-ground",
				position = actual_position,
				stack = item
			}
		end
	elseif remove_from_ground then
		for _, item in ipairs(surface.find_entities_filtered{area = tile_bb, name = "item-on-ground"}) do
			item.destroy()
		end
	end
	return nil
end

-- Callback of the on_init and on_load events. Generates the necessary data if they don't exist.
function on_init_or_load()
	-- Generate event IDs if they are not generated.
	get_or_load_on_enabled_event_id()
	
	-- Generate table for storing data that will be used later.
	if not global.creative_mode then
		global.creative_mode = {}
		-- Store whether creative mode has been enabled.
		global.creative_mode.enabled = false
		-- Sub-tables for storing the entities or data about the entities so we can update them later.
		if not global.creative_mode.item_source_data then global.creative_mode.item_source_data = {} end
		if not global.creative_mode.item_void then global.creative_mode.item_void = {} end
		if not global.creative_mode.fluid_source then global.creative_mode.fluid_source = {} end
		if not global.creative_mode.fluid_void then global.creative_mode.fluid_void = {} end
		if not global.creative_mode.energy_source then global.creative_mode.energy_source = {} end
		if not global.creative_mode.energy_void then global.creative_mode.energy_void = {} end
	end
end

-- Callback of the on_built_entity and on_robot_built_entity events. Registers the built entity if it is one of the creative entities.
function on_built_entity(event)
	if event.created_entity.name == creative_mode_defines.names.entities.item_source then
		table.insert(global.creative_mode.item_source_data,
		{
			entity = event.created_entity, -- The item-source entity.
			slot1_inserted_players = nil, -- The players who have been inserted the item in the first slot.
			slot2_inserted_players = nil -- The players who have been inserted the item in the first slot.
		})
	elseif event.created_entity.name == creative_mode_defines.names.entities.item_void then
		table.insert(global.creative_mode.item_void, event.created_entity)
	elseif event.created_entity.name == creative_mode_defines.names.entities.fluid_source then
		table.insert(global.creative_mode.fluid_source, event.created_entity)
	elseif event.created_entity.name == creative_mode_defines.names.entities.fluid_void then
		table.insert(global.creative_mode.fluid_void, event.created_entity)
	elseif event.created_entity.name == creative_mode_defines.names.entities.energy_source then
		table.insert(global.creative_mode.energy_source, event.created_entity)
	elseif event.created_entity.name == creative_mode_defines.names.entities.energy_void then
		table.insert(global.creative_mode.energy_void, event.created_entity)
	end
end

-- Registers callback for the on_init event, which is invoked when a new game is started.
script.on_init(on_init_or_load)

-- Registers callback for the on_load event, which is invoked when an existing save is loaded.
script.on_load(on_init_or_load)

--[[
-- Registers callback for the on_configuration_changed, which is invoked when the game version or any active mods change.
script.on_configuration_changed(function(data)
	-- Perform changes if our mod version has changed.
	if data.mod_changes then
		local game_old_version = data.old_version
		local game_new_version = data.new_version
		local my_mod_changes = data.mod_changes["creative-mode"]
		if my_mod_changes then
			local old_version = my_mod_changes.old_version
			local new_version = my_mod_changes.new_version
		end
	end
end)
]]--

-- Registers callback for the on_built_entity event.
script.on_event(defines.events.on_built_entity, on_built_entity)

-- Registers callback for the on_robot_built_entity event.
script.on_event(defines.events.on_robot_built_entity, on_built_entity)

-- Ticking.
script.on_event(defines.events.on_tick, function(event)
	-- Loop through the table of item-source data to output items.
	for index, item_source_data in ipairs(global.creative_mode.item_source_data) do
		-- Get the actual item-source entity.
		local item_source = item_source_data.entity
		-- Work only if the entity is valid.
		if item_source.valid then
			-- Check if it is active and also not marked for deconstruction.
			if item_source.active and not item_source.to_be_deconstructed(item_source.force) then
				-- Give the item-source free energy.
				item_source.energy = 100000 -- It seems not working?
				-- Get the item names in the 2 filter slots of the item-source.
				local slot1 = item_source.get_filter(1)
				local slot2 = item_source.get_filter(2)
				-- Get the item-source's surface, position and shift for output, so we can drop items accordingly.
				local surf = item_source.surface
				local pos = item_source.position
				local dir = item_source.direction
				local shift = item_source_shift[dir]
				-- Output for slot1.
				if slot1 == nil then
					item_source_data.slot1_inserted_players = nil
				else
					item_source_data.slot1_inserted_players = output_or_remove_item(surf, pos, shift.x1, shift.y1, dir, true, slot1, 1, item_source_data.slot1_inserted_players)
				end
				-- Output for slot2.
				if slot2 == nil then
					item_source_data.slot2_inserted_players = nil
				else
					item_source_data.slot2_inserted_players = output_or_remove_item(surf, pos, shift.x2, shift.y2, dir, true, slot2, 2, item_source_data.slot2_inserted_players)
				end
			end
		else
			-- Remove invalid entity.
			table.remove(global.creative_mode.item_source_data, index)
		end
	end
	
	-- Loop through the table of item-void to nullify items.
	for index, item_void in ipairs(global.creative_mode.item_void) do
		if item_void.valid then
			-- Check if it is active and also not marked for deconstruction.
			if item_void.active and not item_void.to_be_deconstructed(item_void.force) then
				-- Give the item-void free energy.
				item_void.energy = 100000 -- It seems not working?
				-- Get the item-void's surface, position and shift for item removal.
				local surf = item_void.surface
				local pos = item_void.position
				local dir = item_void.direction
				local shift = item_void_shift[dir]
				-- Remove the items in front of it.
				output_or_remove_item(surf, pos, shift.x, shift.y, dir, false, nil, nil, nil)
			end
		else
			table.remove(global.creative_mode.item_void, index)
		end
	end
	
	-- Loop through the table of fluid-source to give free energy.
	for index, fluid_source in ipairs(global.creative_mode.fluid_source) do
		if fluid_source.valid then
			fluid_source.energy = 100000
		else
			table.remove(global.creative_mode.fluid_source, index)
		end
	end
	
	-- Loop through the table of fluid-void to nullify the fluid inside its fluid box.
	for index, fluid_void in ipairs(global.creative_mode.fluid_void) do
		if fluid_void.valid then
			for i = 1, #fluid_void.fluidbox do
				fluid_void.fluidbox[i] = nil
			end
		else
			table.remove(global.creative_mode.fluid_void, index)
		end
	end
	
	-- Loop through the table of energy-source to give free energy.
	for index, energy_source in ipairs(global.creative_mode.energy_source) do
		if energy_source.valid then
			energy_source.energy = 1000000000000000
		else
			table.remove(global.creative_mode.energy_source, index)
		end
	end
	
	-- Loop through the table of energy-void to nullify the energy.
	for index, energy_void in ipairs(global.creative_mode.energy_void) do
		if energy_void.valid then
			energy_void.energy = 0
			-- Also clear its output inventory, though it is unlikely to finish crafting.
			energy_void.get_output_inventory().clear()
			-- Fix its recipe. Somehow player can copy and paste other recipes from another assembling machine.
			energy_void.recipe = energy_void.force.recipes[creative_mode_defines.names.recipes.energy_absorption]
		else
			table.remove(global.creative_mode.energy_void, index)
		end
	end
end)

-- Registers callback for the on_gui_click event, which is invoked when a GUI element is clicked.
script.on_event(defines.events.on_gui_click, function(event)
	local player = game.players[event.element.player_index]
	if event.element.name == creative_mode_defines.names.gui.enable_cheat_mode_yes then
		-- Enable cheat mode? Yes
		enable_cheat_mode(player)
		player.gui.center.popup.destroy()
		show_popup(2, player)
	elseif event.element.name == creative_mode_defines.names.gui.enable_cheat_mode_no then
		-- Enable cheat mode? No
		player.gui.center.popup.destroy()
		show_popup(2, player)
	elseif event.element.name == creative_mode_defines.names.gui.research_all_tech_yes then
		-- Research all tech? Yes
		research_all_technologies(player)
		player.gui.center.popup.destroy()
		show_popup(3, player)
	elseif event.element.name == creative_mode_defines.names.gui.research_all_tech_no then
		-- Research all tech? No
		player.gui.center.popup.destroy()
		show_popup(3, player)
	elseif event.element.name == creative_mode_defines.names.gui.make_all_surface_always_day_yes then
		-- Always day? Yes
		make_all_surfaces_always_day(player)
		player.gui.center.popup.destroy()
	elseif event.element.name == creative_mode_defines.names.gui.make_all_surface_always_day_no then
		-- Always day? No
		player.gui.center.popup.destroy()
	end
end)

-- Enables cheat mode for the given player.
function enable_cheat_mode(player)
	player.cheat_mode = true
	player.print({"text.creative-mode_cheat-mode-enabled"})
end

-- Researches all technologies for the given player.
function research_all_technologies(player)
	player.force.research_all_technologies()
	player.print({"text.creative-mode_all-technologies-researched"})
end

-- Makes all surfaces always day.
function make_all_surfaces_always_day(player)
	for _, surface in pairs(game.surfaces) do
		surface.always_day = true
	end
	player.print({"text.creative-mode_made-all-surfaces-always-day"})
end

-- Show a popup to ask for decision according to the given stage number.
function show_popup(stage, player)
	local center = player.gui.center
	if stage == 1 then
		-- Enable cheat mode?
		center.add({type = "frame", direction = "horizontal", name = "popup", caption = {"text.creative-mode_ask-enable-cheat-mode"}})
		center.popup.add({type = "button", name = creative_mode_defines.names.gui.enable_cheat_mode_yes, caption = {"text.creative-mode_yes"}})
		center.popup.add({type = "button", name = creative_mode_defines.names.gui.enable_cheat_mode_no, caption = {"text.creative-mode_no"}})
	elseif stage == 2 then
		-- Research all tech?
		center.add({type = "frame", direction = "horizontal", name = "popup", caption = {"text.creative-mode_ask-research-all-technologies"}})
		center.popup.add({type = "button", name = creative_mode_defines.names.gui.research_all_tech_yes, caption = {"text.creative-mode_yes"}})
		center.popup.add({type = "button", name = creative_mode_defines.names.gui.research_all_tech_no, caption = {"text.creative-mode_no"}})
	elseif stage == 3 then
		-- Always day?
		center.add({type = "frame", direction = "horizontal", name = "popup", caption = {"text.creative-mode_ask-make-all-surfaces-always-day"}})
		center.popup.add({type = "button", name = creative_mode_defines.names.gui.make_all_surface_always_day_yes, caption = {"text.creative-mode_yes"}})
		center.popup.add({type = "button", name = creative_mode_defines.names.gui.make_all_surface_always_day_no, caption = {"text.creative-mode_no"}})
	end
end

-- Create interface for lua command.
remote.add_interface(creative_mode_defines.names.interface,
{
	-- Create a custom event to notify other mods that creative mode is enabled.
	on_enabled = function()
		return get_or_load_on_enabled_event_id()
	end,
	
	-- The command for enabling creative mode.
	enable = function()
		-- If creative mode has been enabled, notify the player and do nothing.
		if global.creative_mode.enabled then
			game.player.print({"text.creative-mode_already-enabled"})
			return
		end
		-- Enable the recipes.
		for _, recipe in pairs(creative_mode_defines.names.recipes) do
			if game.player.force.recipes[recipe] ~= nil then
				game.player.force.recipes[recipe].enabled = true
			end
		end
		-- Enable the free fluid recipes. Get the recipe names by looping through the fluid prototypes in game.
		for _, fluid in pairs(game.fluid_prototypes) do
			game.player.force.recipes[creative_mode_defines.names.free_fluid_recipe_prefix .. fluid.name].enabled = true
		end
		
		-- Also unhide the items so they can be selected as filter.
		-- Unfortunately, LuaItemPrototype doesn't have flags property. This cannot be done at this moment.
--		for _, item in pairs(creative_mode_defines.names.items) do
--			if game.item_prototypes[item] ~= nil then
--				table.remove(game.item_prototypes[item].flags, 2) -- The hidden flag is the second one.
--			end
--		end
		
		-- Increase player's reach.
		if creative_mode_config.misc.also_increase_player_reach then
			-- Check if player has a character.
			if game.player.character then
				game.player.character_build_distance_bonus = 10000000
				game.player.character_item_drop_distance_bonus = 10000000
				game.player.character_reach_distance_bonus = 10000000
				game.player.character_resource_reach_distance_bonus = 10000000
			end
		end
		-- Increase player's mining speed.
		if creative_mode_config.misc.also_increase_player_mining_speed then
			game.player.character_mining_speed_modifier = 100
		end
		
		-- Notify the player.
		game.player.print({"text.creative-mode_enable-succeeded"})
		
		-- Also raise the custom event.
		game.raise_event(on_enabled_event_id, {creative_mode_version = creative_mode_defines.version})
		
		-- Save as creative mode has been enabled.
		global.creative_mode.enabled = true
		
		-- Show popup for enabling more features.
		if creative_mode_config.misc.also_show_popups_for_more_features then
			show_popup(1, game.player)
		end
	end
})